#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <GString.h>
#include <gmem.h>
#include <gfile.h>
#include <config.h>
#include "Object.h"
#include "Stream.h"
#include "Array.h"
#include "Dict.h"
#include "XRef.h"
#include "Catalog.h"
#include "Page.h"
#include "GfxFont.h"
#include "PDFDoc.h"
#include "Params.h"
#include "Error.h"

#include "epdf.h"

#define obj_free(obj)  if (!obj.isNull())    obj.free()

// when copying the Resources of the selected page, all objects are copied
// recusively top-down. Indirect objects however are not fetched during
// copying, but get a new object number from pdftex and then will be
// appended into a linked list. Duplicates are checked and removed from the
// list of indirect objects during appending.

enum InObjType {
    objFont,
    objFontDesc,
    objOther
};

struct InObj {
    Ref ref;            // ref in original PDF
    InObjType type;     // object type
    InObj *next;        // next entry in list of indirect objects
    integer num;        // new object number in output PDF
    int fontmap;        // index of font map entry
    integer encoding;   // Encoding for objFont       
};

struct UsedEncoding {
    integer encoding;
    GfxFont *font;
    UsedEncoding *next;
};

InObj *inObjList;
UsedEncoding *encodingList;
static GBool isInit = gFalse;

static int addEncoding(GfxFont *gfont)
{
    UsedEncoding *n;
    n = new UsedEncoding[1];
    n->next = encodingList;
    encodingList = n;
    n->font = gfont;
    n->encoding = pdfnewobjnum();
    return n->encoding;
}

#define addFont(ref, fontmap, encoding) \
        addInObj(objFont, ref, fontmap, encoding)

#define addFontDesc(ref, fontmap) \
        addInObj(objFontDesc, ref, fontmap, 0)

#define addOther(ref) \
        addInObj(objOther, ref, 0, 0)

static int addInObj(InObjType type, Ref ref, int f, integer e)
{
    InObj *p, *q, *n = new InObj[1];
    if (ref.num == 0)
        pdftex_fail("pdf inclusion: invalid reference");
    n->ref = ref;
    n->type = type;
    n->next = 0;
    n->fontmap = f;
    n->encoding = e;
    if (inObjList == 0)
        inObjList = n;
    else {
        for (p = inObjList; p != 0; p = p->next) {
            if (p->ref.num == ref.num && p->ref.gen == ref.gen) {
                delete n;
                return p->num;
            }
            q = p;
        }
        q->next = n;
    }
    n->num = pdfnewobjnum();
    return n->num;
}

static void copyObject(Object *);

static void copyName(char *s)
{
    pdf_puts("/");
    for (; *s != 0; s++) {
        if (isdigit(*s) || isupper(*s) || islower(*s) || *s == '_')
            pdfout(*s);
        else
            pdf_printf("#%.2X", *s & 0xFF);
    }
}

static void copyDictEntry(Object *obj, int i)
{
    Object obj1;
    copyName(obj->dictGetKey(i));
    pdf_puts(" ");
    obj->dictGetValNF(i, &obj1);
    copyObject(&obj1);
    obj1.free();
    pdf_puts("\n");
}

static void copyDict(Object *obj)
{
    int i, l;
    if (!obj->isDict())
        pdftex_fail("pdf inclusion: invalid dict type <%s>", 
                    obj->getTypeName());
    for (i = 0, l = obj->dictGetLength(); i < l; ++i)
        copyDictEntry(obj, i);
}

static void copyFontDict(Object *obj, InObj *r)
{
    int i, k, l;
    char *key;
    if (!obj->isDict())
        pdftex_fail("pdf inclusion: invalid dict type <%s>", 
                    obj->getTypeName());
    pdf_puts("<<\n");
    if (r->type == objFont) { // Font dict
        for (i = 0, l = obj->dictGetLength(); i < l; ++i) {
            key = obj->dictGetKey(i);
            if (strcmp("BaseFont", key) == 0 || 
                strcmp("Encoding", key) == 0)
                continue; // skip original values
            copyDictEntry(obj, i);
        }
        // write new BaseFont and Encoding
        pdf_printf("/BaseFont %i 0 R\n", (int)get_fontname(r->fontmap)); 
        pdf_printf("/Encoding %i 0 R\n", (int)r->encoding); 
    }
    else { // FontDescriptor dict
        for (i = 0, l = obj->dictGetLength(); i < l; ++i) {
            key = obj->dictGetKey(i);
            if (strcmp("FontName", key) == 0 ||
                strncmp("FontFile", key, strlen("FontFile")) == 0)
                continue; // ignore original FontFile/FontName
            if (strcmp("CharSet", key) == 0)
                continue; // ignore CharSet
            copyDictEntry(obj, i);
        }
        // write new FontName and FontFile
        pdf_printf("/FontName %i 0 R\n", (int)get_fontname(r->fontmap)); 
        pdf_printf("/FontFile %i 0 R\n", (int)get_fontfile(r->fontmap));
    }
    pdf_puts(">>");
}

static void copyStream(Stream *str)
{
    int c;
    str->reset();
    while ((c = str->getChar()) != EOF)
        pdfout(c);
}

static void copyProcSet(Object *obj)
{
    int i, l;
    Object procset;
    if (!obj->isArray())
        pdftex_fail("pdf inclusion: invalid ProcSet array type <%s>", 
                    obj->getTypeName());
    pdf_puts("/ProcSet [ ");
    for (i = 0, l = obj->arrayGetLength(); i < l; ++i) {
        obj->arrayGet(i, &procset);
        if (!procset.isName())
            pdftex_fail("pdf inclusion: invalid ProcSet entry type <%s>", 
                        procset.getTypeName());
        copyName(procset.getName());
        pdf_puts(" ");
        procset.free();
    }
    pdf_puts("]\n");
}

static void copyFont(char *tag, Object *fontRef)
{
    Object fontdict, subtype, basefont, fontdescRef, fontdesc, charset;
    GfxFont *gfont;
    int fontmap;
    fontdict.initNull();
    subtype.initNull();
    basefont.initNull();
    fontdescRef.initNull();
    fontdesc.initNull();
    fontRef->fetch(&fontdict);
    if (!fontdict.isDict())
        pdftex_fail("pdf inclusion: invalid font dict type <%s>", 
                    fontdict.getTypeName());
    fontdict.dictLookup("Subtype", &subtype);
    if (!subtype.isName())
        pdftex_fail("pdf inclusion: invalid font Subtype entry type <%s>", 
                    subtype.getTypeName());
    /* only handle Type1 fonts; others will be copied */
    if (strcmp(subtype.getName(), "Type1") != 0 ) {
            copyName(tag);
            pdf_puts(" ");
            copyObject(fontRef);
            goto free_objects;
    }
    fontdict.dictLookup("BaseFont", &basefont);
    if (!basefont.isName())
        pdftex_fail("pdf inclusion: invalid font BaseFont entry type <%s>", 
                    basefont.getTypeName());
    fontmap = lookup_fontmap(basefont.getName());
    if (fontmap >= 0 && is_type1(fontmap) &&
        fontdict.dictLookupNF("FontDescriptor", &fontdescRef) && 
        fontdescRef.isRef() && fontdescRef.fetch(&fontdesc)) {
        if (fontdesc.dictLookup("CharSet", &charset) && 
            charset.isString() && is_subsetable(fontmap)) {
            mark_glyphs(fontmap, charset.getString()->getCString());
            charset.free();
        }
        else
            embed_whole_font(fontmap);
        addFontDesc(fontdescRef.getRef(), fontmap);
    }
    copyName(tag);
    if (!fontdesc.isNull()) {
        gfont = new GfxFont(tag, fontRef->getRef(), fontdict.getDict());
        pdf_printf(" %d 0 R ", addFont(fontRef->getRef(), fontmap, 
                                       addEncoding(gfont)));
    }
    else
        pdf_printf(" %d 0 R ", addOther(fontRef->getRef()));
free_objects:
    obj_free(fontdict);
    obj_free(subtype);
    obj_free(basefont);
    obj_free(fontdescRef);
    obj_free(fontdesc);
}

static void copyFontResources(Object *obj)
{
    Object fontRef;
    int i, l;
    if (!obj->isDict())
        pdftex_fail("pdf inclusion: invalid font resources dict type <%s>", 
                    obj->getTypeName());
    pdf_puts("/Font << ");
    for (i = 0, l = obj->dictGetLength(); i < l; ++i) {
        obj->dictGetValNF(i, &fontRef);
        if (fontRef.isRef())
            copyFont(obj->dictGetKey(i), &fontRef);
        else
            pdftex_fail("pdf inclusion: invalid font in reference type <%s>", 
                        fontRef.getTypeName());
        fontRef.free();
    }
    pdf_puts(">>\n");
}

static void copyOtherResources(Object *obj, char *key)
{
    Object obj1;
    int i, l;
    if (!obj->isDict())
        pdftex_fail("pdf inclusion: invalid other resources dict type <%s>", 
                    obj->getTypeName());
    copyName(key);
    pdf_puts(" << ");
    for (i = 0, l = obj->dictGetLength(); i < l; ++i) {
        obj->dictGetValNF(i, &obj1);
        if (obj1.isRef()) {
            copyName(obj->dictGetKey(i));
            pdf_printf(" %d 0 R ", addOther(obj1.getRef()));
        }
        else 
            copyObject(&obj1);
        obj1.free();
    }
    pdf_puts(">>\n");
}

static void copyObject(Object *obj)
{
    Object obj1;
    int  i, l, c;
    Ref ref;
    char *p;
    GString *s;
    if (obj->isBool()) {
        pdf_printf("%s", obj->getBool() ? "true" : "false");
    }
    else if (obj->isInt()) {
        pdf_printf("%i", obj->getInt());
    }
    else if (obj->isReal()) {
        pdf_printf("%g", obj->getReal());
    }
    else if (obj->isNum()) {
        pdf_printf("%g", obj->getNum());
    }
    else if (obj->isString()) {
        s = obj->getString();
        p = s->getCString();
        l = s->getLength();
        if (strlen(p) == l) {
            pdf_puts("(");
            for (; *p != 0; p++) {
                c = *p;
                if (c == '(' || c == ')' || c == '\\')
                    pdf_printf("\\%c", c);
                else if (c < 0x20 || c > 0x7F)
                    pdf_printf("\\%03o", c);
                else
                    pdfout(c);
            }
            pdf_puts(")");
        }
        else {
            pdf_puts("<");
            for (i = 0; i < l; i++) {
                c = s->getChar(i) & 0xFF;
                pdf_printf("%.2x", c);
            }
            pdf_puts(">");
        }
    }
    else if (obj->isName()) {
        copyName(obj->getName());
    }
    else if (obj->isNull()) {
        pdf_puts("null");
    }
    else if (obj->isArray()) {
        pdf_puts("[");
        for (i = 0, l = obj->arrayGetLength(); i < l; ++i) {
            obj->arrayGetNF(i, &obj1);
            if (!obj1.isName())
                pdf_puts(" ");
            copyObject(&obj1);
            obj1.free();
        }
        pdf_puts("]");
    }
    else if (obj->isDict()) {
        pdf_puts("<<\n");
        copyDict(obj);
        pdf_puts(">>");
    }
    else if (obj->isStream()) {
        obj1.initDict(obj->getStream()->getDict());
        obj->getStream()->getDict()->incRef();
        pdf_puts("<<\n");
        copyDict(&obj1);
        obj1.free();
        pdf_puts(">>\n");
        pdf_puts("stream\n");
        copyStream(obj->getStream()->getBaseStream());
        pdf_puts("endstream");
    }
    else if (obj->isRef()) {
        ref = obj->getRef();
        if (ref.num == 0) {
            pdftex_warn("pdf inclusion: reference to invalid object was replaced by <null>");
            pdf_puts("null");
        }
        else
            pdf_printf("%d 0 R", addOther(ref));
    }
    else {
        pdftex_fail("pdf inclusion: type <%s> cannot be copied", 
                    obj->getTypeName());
    }
}

static void writeRefs()
{
    Object obj1;
    InObj *r, *n;
    for (r = inObjList; r != 0; r = r->next) {
        zpdfbeginobj(r->num);
        xref->fetch(r->ref.num, r->ref.gen, &obj1);
        if (r->type == objFont || r->type == objFontDesc)
            copyFontDict(&obj1, r);
        else
            copyObject(&obj1);
        pdf_puts("\n");
        pdfendobj();
        obj1.free();
    }
    for (r = inObjList; r != 0; r = n) {
        n = r->next;
        delete r;
    }
}

static void writeEncodings()
{
    UsedEncoding *n, *r;
    char *glyphNames[MAX_CHAR_CODE + 1], *s;
    int i;
    for (r = encodingList; r != 0; r = r->next) {
        for (i = 0; i <= MAX_CHAR_CODE; i++)
            if ((s = r->font->getCharName(i)) != 0)
                glyphNames[i] = s;
            else
                glyphNames[i] = notdef;
        write_enc(glyphNames, r->encoding);
    }
    for (r = encodingList; r != 0; r = n) {
        n = r->next;
        delete r->font;
        delete r;
    }
}

integer read_pdf_info(char *image_name, integer page_num)
{
    PDFDoc *doc;
    GString *docName;
    Page *page;
    // initialize
    if (!isInit) {
        initParams(xpdfConfigFile);
        errorInit();
        isInit = gTrue;
    }
    // open PDF file
    xref = 0;
    docName = new GString(image_name);
    doc = new PDFDoc(docName);
    if (!doc->isOk() || !doc->okToPrint())
        pdftex_fail("xpdf: reading PDF image failed");
    epdf_doc = (void *)doc;
    epdf_xref = (void *)xref;
    epdf_num_pages = doc->getCatalog()->getNumPages();
    if (page_num <= 0 || page_num > epdf_num_pages)
        pdftex_fail("pdf inclusion: required page does not exists <%i>", 
                    epdf_num_pages);
    // get the required page
    page = doc->getCatalog()->getPage(page_num);
    if (page->isCropped()) {
        epdf_width = page->getCropX2() - page->getCropX1();
        epdf_height = page->getCropY2() - page->getCropY1();
        epdf_orig_x = page->getCropX1();
        epdf_orig_y = page->getCropY1();
    }
    else {
        epdf_width = page->getX2() - page->getX1();
        epdf_height = page->getY2() - page->getY1();
        epdf_orig_x = page->getX1();
        epdf_orig_y = page->getY1();
    }
    return 0;
}

void write_epdf(void)
{
    Page *page;
    Object contents, obj1, obj2;
    char *key;
    int i, l;
    xref = (XRef *)epdf_xref;
    page = ((PDFDoc *)epdf_doc)->getCatalog()->getPage(epdf_selected_page);
    inObjList = 0;
    encodingList = 0;
    // write the Page header
    pdf_puts("/Type /XObject\n");
    pdf_puts("/Subtype /Form\n");
    pdf_puts("/FormType 1\n");
    pdf_puts("/Matrix [1 0 0 1 0 0]\n");
    if (page->isCropped())
        pdf_printf("/BBox [%.5g %.5g %.5g %.5g]\n",
                  page->getCropX1(),
                  page->getCropY1(),
                  page->getCropX2(),
                  page->getCropY2());
    else 
        pdf_printf("/BBox [%.5g %.5g %.5g %.5g]\n",
                  page->getX1(),
                  page->getY1(),
                  page->getX2(),
                  page->getY2());
    // write the Resources dictionary
    obj1.initDict(page->getResourceDict());
    page->getResourceDict()->incRef();
    if (!obj1.isDict())
        pdftex_fail("pdf inclusion: invalid resources dict type <%s>", 
                    obj1.getTypeName());
    pdf_puts("/Resources <<\n");
    for (i = 0, l = obj1.dictGetLength(); i < l; ++i) {
        obj1.dictGetVal(i, &obj2);
        key = obj1.dictGetKey(i);
        if (strcmp("Font", key) == 0)
            copyFontResources(&obj2);
        else if (strcmp("ProcSet", key) == 0)
            copyProcSet(&obj2);
        else
            copyOtherResources(&obj2, key);
        obj2.free();
    }
    obj1.free();
    pdf_puts(">>\n");
    // write the page contents
    page->getContents(&contents);
    if (contents.isStream()) {
        obj1.initDict(contents.getStream()->getDict());
        contents.getStream()->getDict()->incRef();
        copyDict(&obj1);
        pdf_puts(">>\nstream\n");
        copyStream(contents.getStream()->getBaseStream());
        pdf_puts("endstream\n");
        pdfendobj();
        obj1.free();
    }
    else if (contents.isArray()) {
        pdfbeginstream();
        for (i = 0, l = contents.arrayGetLength(); i < l; ++i) {
            contents.arrayGet(i, &obj1);
            copyStream(obj1.getStream());
            obj1.free();
        }
        pdfendstream();
    }
    else
        pdftex_fail("pdf inclusion: invalid page contents type <%s>", 
                    contents.getTypeName());
    contents.free();
    // write out all indirect objects
    writeRefs();
    // write out all used encodings
    writeEncodings();
}

void epdf_delete()
{
    xref = (XRef *)epdf_xref;
    delete (PDFDoc *)epdf_doc;
}

void epdf_check_mem()
{
    if (isInit) {
        freeParams();
        Object::memCheck(errFile);
        gMemReport(errFile);
    }
}
